@@ -1,4 +1,8 @@
-#include "math.h" // math first to fix OSX error
+// This is a personal academic project. Dear PVS-Studio, please check it.
+
+// PVS-Studio Static Code Analyzer for C, C++, C#, and Java: https://pvs-studio.com
+
+#include "mathtools.h" // math first to fix OSX error
 #include "stamp.h"
 #include "fft.h"
 #include "mass.h"
@@ -9,6 +13,7 @@
 #include <RcppThread.h>
 // [[Rcpp::depends(RcppParallel)]]
 #include <RcppParallel.h>
+#include <cmath>
 using namespace RcppParallel;
 
 #if RCPP_PARALLEL_USE_TBB
@@ -23,18 +28,17 @@ List stamp_rcpp(const NumericVector data_ref, const NumericVector query_ref, uin
                 double s_size, bool progress) {
 
   bool partial = false;
-  int64_t exclusion_zone = round(window_size * ez + DBL_EPSILON);
-  uint64_t data_size = data_ref.length();
-  uint64_t query_size = query_ref.length();
-  uint64_t matrix_profile_size = data_size - window_size + 1;
-  uint64_t num_queries = query_size - window_size + 1;
+  int64_t const exclusion_zone = round(window_size * ez + DBL_EPSILON);
+  uint64_t const data_size = data_ref.length();
+  uint64_t const query_size = query_ref.length();
+  uint64_t const matrix_profile_size = data_size - window_size + 1;
+  uint64_t const num_queries = query_size - window_size + 1;
 
   // check skip position
   LogicalVector skip_location(matrix_profile_size);
 
-  // TODO: data or query?
   for (uint64_t i = 0; i < matrix_profile_size; i++) {
-    NumericVector range = data_ref[Range(i, (i + window_size - 1))];
+    NumericVector const range = data_ref[Range(i, (i + window_size - 1))];
     if (any(is_na(range) | is_infinite(range))) {
       skip_location[i] = TRUE;
     }
@@ -54,7 +58,7 @@ List stamp_rcpp(const NumericVector data_ref, const NumericVector query_ref, uin
 
   IntegerVector order = Range(0, num_queries - 1);
 
-  uint32_t k = find_best_k_rcpp(data, query, window_size);
+  uint32_t const grain = find_best_k_rcpp(data, query, window_size);
 
   Progress p(num_queries, progress);
 
@@ -68,32 +72,32 @@ List stamp_rcpp(const NumericVector data_ref, const NumericVector query_ref, uin
 
   try {
     uint64_t j = 1;
-    for (int32_t i : order) {
+    for (int32_t const i : order) {
       RcppThread::checkUserInterrupt();
       p.increment();
 
-      List nn =
-          mass3_rcpp(query[Range(i, i + window_size - 1)], data, pre["data_size"], pre["window_size"], pre["data_mean"],
-                     pre["data_sd"], as<NumericVector>(pre["query_mean"])[i], as<NumericVector>(pre["query_sd"])[i], k);
+      List nn = mass3_rcpp(query[Range(i, i + window_size - 1)], data, pre["data_size"], pre["window_size"],
+                           pre["data_mean"], pre["data_sd"], as<NumericVector>(pre["query_mean"])[i],
+                           as<NumericVector>(pre["query_sd"])[i], grain);
 
       NumericVector distance_profile = as<NumericVector>(nn["distance_profile"]);
 
       // apply exclusion zone
       if (exclusion_zone > 0) {
-        int64_t exc_st = MAX(0, (i - exclusion_zone));
-        int64_t exc_ed = MIN(matrix_profile_size - 1, (uint64_t)(i + exclusion_zone));
-        IntegerVector dp_range = Range(exc_st, exc_ed);
+        int64_t const exc_st = MAX(0, (i - exclusion_zone));
+        int64_t const exc_ed = MIN(matrix_profile_size - 1, (uint64_t)(i + exclusion_zone));
+        IntegerVector const dp_range = Range(exc_st, exc_ed);
         distance_profile[dp_range] = R_PosInf;
       }
 
       distance_profile[as<NumericVector>(pre["data_sd"]) < DBL_EPSILON] = R_PosInf;
-      if (skip_location[i] || as<NumericVector>(pre["query_sd"])[i] < DBL_EPSILON) {
+      if ((skip_location[i] != 0) || as<NumericVector>(pre["query_sd"])[i] < DBL_EPSILON) {
         distance_profile.fill(R_PosInf);
       }
       distance_profile[skip_location] = R_PosInf;
 
       // normal matrix_profile
-      LogicalVector idx = (distance_profile < matrix_profile);
+      LogicalVector const idx = (distance_profile < matrix_profile);
       matrix_profile[idx] = distance_profile[idx];
       profile_index[which_cpp(idx)] = i + 1;
 
@@ -109,7 +113,7 @@ List stamp_rcpp(const NumericVector data_ref, const NumericVector query_ref, uin
     partial = true;
     Rcout << "Process terminated by the user successfully, partial results were returned." << std::endl;
   } catch (...) {
-    ::Rf_error("c++ exception (unknown reason)");
+    Rcpp::stop("c++ exception (unknown reason)");
   }
 
   return (List::create(Rcpp::Named("matrix_profile") = sqrt(matrix_profile),
@@ -118,6 +122,7 @@ List stamp_rcpp(const NumericVector data_ref, const NumericVector query_ref, uin
 }
 
 struct StampWorker : public Worker {
+private:
   // input
   const RVector<double> data_ref;
   const RVector<double> window_ref;
@@ -140,26 +145,26 @@ struct StampWorker : public Worker {
 #else
   tthread::mutex m;
 #endif
-
+public:
   // initialize from Rcpp input and output matrixes (the RMatrix class
   // can be automatically converted to from the Rcpp matrix type)
-  StampWorker(const NumericVector data_ref, const NumericVector window_ref, const uint64_t w_size,
-              const uint64_t d_size, const NumericVector d_mean, const NumericVector d_std, const NumericVector q_mean,
-              const NumericVector q_std, const IntegerVector skip_location, const uint64_t ez, Progress *p,
-              NumericVector mp, IntegerVector pi)
+  StampWorker(const NumericVector &data_ref, const NumericVector &window_ref, const uint64_t w_size,
+              const uint64_t d_size, const NumericVector &d_mean, const NumericVector &d_std,
+              const NumericVector &q_mean, const NumericVector &q_std, const IntegerVector &skip_location,
+              const uint64_t ez, Progress *p, const NumericVector &mp, const IntegerVector &pi)
       : data_ref(data_ref), window_ref(window_ref), w_size(w_size), d_size(d_size), d_mean(d_mean), d_std(d_std),
         q_mean(q_mean), q_std(q_std), skip_location(skip_location), ez(ez), p(p), mp(mp), pi(pi) {}
 
-  ~StampWorker() {}
+  ~StampWorker() override = default;
 
   // function call operator that work for the specified range (begin/end)
-  void operator()(std::size_t begin, std::size_t end) {
+  void operator()(std::size_t begin, std::size_t end) override {
     // begin and end are the indexes of data.
-    uint64_t i, j;
-    uint64_t start_ez, end_ez;
-    double dp;
+    uint64_t i = 0, j = 0;
+    uint64_t start_ez = 0, end_ez = 0;
+    double dp = NAN;
 
-    uint64_t chunk = (end - begin);
+    uint64_t const chunk = (end - begin);
 
     if (chunk <= w_size) {
       Rcout << "Chunk size is too small (" << chunk << ") for a window size of " << w_size << std::endl;
@@ -194,7 +199,7 @@ struct StampWorker : public Worker {
 
         uint64_t jump = end - begin;
         uint64_t k = jump + w_size - 1;
-        uint64_t s = pow(2, (ceil(log2(k))));
+        uint64_t const s = pow(2, (ceil(log2(k))));
 
         std::vector<std::complex<double>> rev_query(s);
 
@@ -255,15 +260,15 @@ struct StampWorker : public Worker {
 List stamp_rcpp_parallel(const NumericVector data_ref, const NumericVector query_ref, uint32_t window_size, double ez,
                          double s_size, bool progress) {
 
-  uint64_t data_size = data_ref.length();
-  uint64_t matrix_profile_size = data_size - window_size + 1;
-  uint64_t exclusion_zone = round(window_size * ez + DBL_EPSILON);
+  uint64_t const data_size = data_ref.length();
+  uint64_t const matrix_profile_size = data_size - window_size + 1;
+  uint64_t const exclusion_zone = round(window_size * ez + DBL_EPSILON);
   bool partial = false;
 
   IntegerVector skip_location(matrix_profile_size, 0);
 
   for (uint64_t i = 0; i < matrix_profile_size; i++) {
-    NumericVector range = data_ref[Range(i, (i + window_size - 1))];
+    NumericVector const range = data_ref[Range(i, (i + window_size - 1))];
     if (any(is_na(range) | is_infinite(range))) {
       skip_location[i] = 1;
     }
@@ -277,14 +282,14 @@ List stamp_rcpp_parallel(const NumericVector data_ref, const NumericVector query
   query[is_na(query)] = 0;
   query[is_infinite(query)] = 0;
 
-  NumericVector matrix_profile(matrix_profile_size, R_PosInf);
-  IntegerVector profile_index(matrix_profile_size, -1);
+  NumericVector const matrix_profile(matrix_profile_size, R_PosInf);
+  IntegerVector const profile_index(matrix_profile_size, -1);
   List pre = mass_pre_rcpp(data, query, window_size);
 
-  uint64_t k = set_k_rcpp(window_size, data_size, window_size);
+  uint64_t const grain = set_k_rcpp(window_size, data_size, window_size);
 
-  uint64_t jobs = pow(2, (ceil(log2((double)data_size / (double)k)) - 1));
-  uint64_t steps = ceil((double)matrix_profile_size / (double)window_size);
+  uint64_t const jobs = pow(2, (ceil(log2((double)data_size / (double)grain)) - 1));
+  uint64_t const steps = ceil((double)matrix_profile_size / (double)window_size);
 
   Progress p(jobs * steps, progress);
 
@@ -295,16 +300,16 @@ List stamp_rcpp_parallel(const NumericVector data_ref, const NumericVector query
   // call parallelFor to do the work
   try {
 #if RCPP_PARALLEL_USE_TBB
-    RcppParallel::parallelFor(0, data.size(), stamp_worker, 2 * k);
+    RcppParallel::parallelFor(0, data.size(), stamp_worker, 2 * grain);
 #else
-    RcppParallel2::ttParallelFor(0, data.size(), stamp_worker, 2 * k);
+    RcppParallel2::ttParallelFor(0, data.size(), stamp_worker, 2 * grain);
 #endif
   } catch (RcppThread::UserInterruptException &e) {
     partial = true;
     Rcout << "Process terminated by the user successfully, partial results "
              "were returned.";
   } catch (...) {
-    ::Rf_error("c++ exception (unknown reason)");
+    Rcpp::stop("c++ exception (unknown reason)");
   }
 
   return (List::create(Rcpp::Named("matrix_profile") = sqrt(matrix_profile),
